Require Import ModalLogic.syntax.
Require Import ModalLogic.semantics.
Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Logic.Classical_Pred_Type.
Require Import Recdef.
Require Import Psatz.
Require Import Coq.Arith.Lt.
Require Import FunInd.
Require Import Coq.Sets.Ensembles.


(* Authors: Cláudia Nalon & Maria Júlia Dias Lima *)

(** The negation normal form takes the usual form: only constants, literals (i.e. propositions or their negations), conjunctions, disjunctions and modal operators are allowed. Note that negatins can only be applied to propositional symbols. The following function returns whether a formula is in NNF. *)

Fixpoint is_NNF (f:formula): Prop :=
  match f with 
  | PropsF p => True
  | ConstF _ => True
  | Not (PropsF p) => True
  | Not _  => False
  | And f g => (is_NNF f ) -> (is_NNF g)
  | Or f g => (is_NNF f ) -> (is_NNF g)
  | Imp _ _ => False
  | Box a f => (is_NNF f)
  | Diamond a f => (is_NNF f)
  end.

(** The next definition takes a formula and returns its NNF. This is defined as a general function, because the recursion is not applied to subformulae. For instance, when taking the transformation of an implication, one of the resulting disjuncts is the negation of its antecedent, which is not a subformula of the implication. Thus, we cannot use structural induction when writing proofs about the NNF a formula. However, the recursion (and, later, our proofs) takes smaller arguments than the formulae it is applied. We take, therefore, the size of a formula as a measure for termination. *)

Function NNF (f:formula) {measure size}: formula :=
  match f with
  | PropsF p => PropsF p
  | ConstF c => ConstF c
  | Not (PropsF p) => Not (PropsF p)
  | Not (ConstF ff) => ConstF tt
  | Not (ConstF tt) => ConstF ff
  | Not (Not f) => NNF f
  | Not (And f g) => Or (NNF (Not f)) (NNF (Not g))
  | Not (Or f g) => And (NNF (Not f)) (NNF (Not g))
  | Not (Imp f g) => And (NNF f) (NNF (Not g))
  | Not (Box a f) => Diamond a (NNF (Not f))
  | Not (Diamond a f) => Box a (NNF (Not f)) 
  | And f g => And (NNF f) (NNF g)
  | Or f g => Or (NNF f) (NNF g)
  | Imp f g => Or (NNF (Not f)) (NNF g)
  | Box a f => Box a (NNF f)
  | Diamond a f => Diamond a (NNF f)
  end.
Proof.
all: try (intros;simpl;lia).
intros. simpl. apply lt_n_S. destruct g; simpl; lia.
Defined.


(* All the admitted theorems below were extracted from Coq Version 8.11.2 *)

Definition R_NNF_correct: forall f _res : formula, _res = NNF f -> R_NNF f _res. 
Admitted. 
Definition R_NNF_complete: forall f _res : formula, R_NNF f _res -> _res = NNF f. 
Admitted.
Definition NNF_rec:
  forall P : formula -> formula -> Set,
  (forall (f : formula) (p : Props), f = PropsF p -> P (PropsF p) (PropsF p)) ->
  (forall f0 : formula, f0 = ConstF ff -> P (ConstF ff) (ConstF ff)) ->
  (forall f1 : formula, f1 = ConstF tt -> P (ConstF tt) (ConstF tt)) ->
  (forall (f2 : formula) (p : Props), f2 = Not (PropsF p) -> P (Not (PropsF p)) (Not (PropsF p))) ->
  (forall f3 : formula, f3 = Not (ConstF ff) -> P (Not (ConstF ff)) (ConstF tt)) ->
  (forall f4 : formula, f4 = Not (ConstF tt) -> P (Not (ConstF tt)) (ConstF ff)) ->
  (forall f5 f6 : formula, f5 = Not (Not f6) -> P f6 (NNF f6) -> P (Not (Not f6)) (NNF f6)) ->
  (forall f6 f7 g : formula,
   f6 = Not (And f7 g) ->
   P (Not f7) (NNF (Not f7)) -> P (Not g) (NNF (Not g)) -> P (Not (And f7 g)) (Or (NNF (Not f7)) (NNF (Not g)))) ->
  (forall f7 f8 g : formula,
   f7 = Not (Or f8 g) ->
   P (Not f8) (NNF (Not f8)) -> P (Not g) (NNF (Not g)) -> P (Not (Or f8 g)) (And (NNF (Not f8)) (NNF (Not g)))) ->
  (forall f8 f9 g : formula,
   f8 = Not (Imp f9 g) -> P f9 (NNF f9) -> P (Not g) (NNF (Not g)) -> P (Not (Imp f9 g)) (And (NNF f9) (NNF (Not g)))) ->
  (forall (f9 : formula) (a : Agents) (f10 : formula),
   f9 = Not (Box a f10) -> P (Not f10) (NNF (Not f10)) -> P (Not (Box a f10)) (Diamond a (NNF (Not f10)))) ->
  (forall (f10 : formula) (a : Agents) (f11 : formula),
   f10 = Not (Diamond a f11) -> P (Not f11) (NNF (Not f11)) -> P (Not (Diamond a f11)) (Box a (NNF (Not f11)))) ->
  (forall f11 f12 g : formula, f11 = And f12 g -> P f12 (NNF f12) -> P g (NNF g) -> P (And f12 g) (And (NNF f12) (NNF g))) ->
  (forall f12 f13 g : formula, f12 = Or f13 g -> P f13 (NNF f13) -> P g (NNF g) -> P (Or f13 g) (Or (NNF f13) (NNF g))) ->
  (forall f13 f14 g : formula,
   f13 = Imp f14 g -> P (Not f14) (NNF (Not f14)) -> P g (NNF g) -> P (Imp f14 g) (Or (NNF (Not f14)) (NNF g))) ->
  (forall (f14 : formula) (a : Agents) (f15 : formula), f14 = Box a f15 -> P f15 (NNF f15) -> P (Box a f15) (Box a (NNF f15))) ->
  (forall (f15 : formula) (a : Agents) (f16 : formula),
   f15 = Diamond a f16 -> P f16 (NNF f16) -> P (Diamond a f16) (Diamond a (NNF f16))) -> forall f16 : formula, P f16 (NNF f16).
Admitted.
Definition NNF_rect:
  forall P : formula -> formula -> Type,
  (forall (f : formula) (p : Props), f = PropsF p -> P (PropsF p) (PropsF p)) ->
  (forall f0 : formula, f0 = ConstF ff -> P (ConstF ff) (ConstF ff)) ->
  (forall f1 : formula, f1 = ConstF tt -> P (ConstF tt) (ConstF tt)) ->
  (forall (f2 : formula) (p : Props), f2 = Not (PropsF p) -> P (Not (PropsF p)) (Not (PropsF p))) ->
  (forall f3 : formula, f3 = Not (ConstF ff) -> P (Not (ConstF ff)) (ConstF tt)) ->
  (forall f4 : formula, f4 = Not (ConstF tt) -> P (Not (ConstF tt)) (ConstF ff)) ->
  (forall f5 f6 : formula, f5 = Not (Not f6) -> P f6 (NNF f6) -> P (Not (Not f6)) (NNF f6)) ->
  (forall f6 f7 g : formula,
   f6 = Not (And f7 g) ->
   P (Not f7) (NNF (Not f7)) -> P (Not g) (NNF (Not g)) -> P (Not (And f7 g)) (Or (NNF (Not f7)) (NNF (Not g)))) ->
  (forall f7 f8 g : formula,
   f7 = Not (Or f8 g) ->
   P (Not f8) (NNF (Not f8)) -> P (Not g) (NNF (Not g)) -> P (Not (Or f8 g)) (And (NNF (Not f8)) (NNF (Not g)))) ->
  (forall f8 f9 g : formula,
   f8 = Not (Imp f9 g) -> P f9 (NNF f9) -> P (Not g) (NNF (Not g)) -> P (Not (Imp f9 g)) (And (NNF f9) (NNF (Not g)))) ->
  (forall (f9 : formula) (a : Agents) (f10 : formula),
   f9 = Not (Box a f10) -> P (Not f10) (NNF (Not f10)) -> P (Not (Box a f10)) (Diamond a (NNF (Not f10)))) ->
  (forall (f10 : formula) (a : Agents) (f11 : formula),
   f10 = Not (Diamond a f11) -> P (Not f11) (NNF (Not f11)) -> P (Not (Diamond a f11)) (Box a (NNF (Not f11)))) ->
  (forall f11 f12 g : formula, f11 = And f12 g -> P f12 (NNF f12) -> P g (NNF g) -> P (And f12 g) (And (NNF f12) (NNF g))) ->
  (forall f12 f13 g : formula, f12 = Or f13 g -> P f13 (NNF f13) -> P g (NNF g) -> P (Or f13 g) (Or (NNF f13) (NNF g))) ->
  (forall f13 f14 g : formula,
   f13 = Imp f14 g -> P (Not f14) (NNF (Not f14)) -> P g (NNF g) -> P (Imp f14 g) (Or (NNF (Not f14)) (NNF g))) ->
  (forall (f14 : formula) (a : Agents) (f15 : formula), f14 = Box a f15 -> P f15 (NNF f15) -> P (Box a f15) (Box a (NNF f15))) ->
  (forall (f15 : formula) (a : Agents) (f16 : formula),
   f15 = Diamond a f16 -> P f16 (NNF f16) -> P (Diamond a f16) (Diamond a (NNF f16))) -> forall f16 : formula, P f16 (NNF f16).
Admitted.
Definition NNF_ind:
  forall P : formula -> formula -> Prop,
  (forall (f : formula) (p : Props), f = PropsF p -> P (PropsF p) (PropsF p)) ->
  (forall f0 : formula, f0 = ConstF ff -> P (ConstF ff) (ConstF ff)) ->
  (forall f1 : formula, f1 = ConstF tt -> P (ConstF tt) (ConstF tt)) ->
  (forall (f2 : formula) (p : Props), f2 = Not (PropsF p) -> P (Not (PropsF p)) (Not (PropsF p))) ->
  (forall f3 : formula, f3 = Not (ConstF ff) -> P (Not (ConstF ff)) (ConstF tt)) ->
  (forall f4 : formula, f4 = Not (ConstF tt) -> P (Not (ConstF tt)) (ConstF ff)) ->
  (forall f5 f6 : formula, f5 = Not (Not f6) -> P f6 (NNF f6) -> P (Not (Not f6)) (NNF f6)) ->
  (forall f6 f7 g : formula,
   f6 = Not (And f7 g) ->
   P (Not f7) (NNF (Not f7)) -> P (Not g) (NNF (Not g)) -> P (Not (And f7 g)) (Or (NNF (Not f7)) (NNF (Not g)))) ->
  (forall f7 f8 g : formula,
   f7 = Not (Or f8 g) ->
   P (Not f8) (NNF (Not f8)) -> P (Not g) (NNF (Not g)) -> P (Not (Or f8 g)) (And (NNF (Not f8)) (NNF (Not g)))) ->
  (forall f8 f9 g : formula,
   f8 = Not (Imp f9 g) -> P f9 (NNF f9) -> P (Not g) (NNF (Not g)) -> P (Not (Imp f9 g)) (And (NNF f9) (NNF (Not g)))) ->
  (forall (f9 : formula) (a : Agents) (f10 : formula),
   f9 = Not (Box a f10) -> P (Not f10) (NNF (Not f10)) -> P (Not (Box a f10)) (Diamond a (NNF (Not f10)))) ->
  (forall (f10 : formula) (a : Agents) (f11 : formula),
   f10 = Not (Diamond a f11) -> P (Not f11) (NNF (Not f11)) -> P (Not (Diamond a f11)) (Box a (NNF (Not f11)))) ->
  (forall f11 f12 g : formula, f11 = And f12 g -> P f12 (NNF f12) -> P g (NNF g) -> P (And f12 g) (And (NNF f12) (NNF g))) ->
  (forall f12 f13 g : formula, f12 = Or f13 g -> P f13 (NNF f13) -> P g (NNF g) -> P (Or f13 g) (Or (NNF f13) (NNF g))) ->
  (forall f13 f14 g : formula,
   f13 = Imp f14 g -> P (Not f14) (NNF (Not f14)) -> P g (NNF g) -> P (Imp f14 g) (Or (NNF (Not f14)) (NNF g))) ->
  (forall (f14 : formula) (a : Agents) (f15 : formula), f14 = Box a f15 -> P f15 (NNF f15) -> P (Box a f15) (Box a (NNF f15))) ->
  (forall (f15 : formula) (a : Agents) (f16 : formula),
   f15 = Diamond a f16 -> P f16 (NNF f16) -> P (Diamond a f16) (Diamond a (NNF f16))) -> forall f16 : formula, P f16 (NNF f16).
Admitted.


(** The proof that the above function is indeed terminating basically relies on properties of inequalities and are automatically obtained via the procedures implemented in coq for linear arithmetic. *)

(** The first theorem shows that the transformationn into the normal form is correct, that is, that the function does produce a formula into the desired formula. *)

Theorem NNF_is_NNF (f:formula) : 
  is_NNF (NNF f).
Proof.
  functional induction (NNF f) using NNF_ind.
  all: (simpl;auto).
Qed.

Theorem eq_NNF_f (A:Ensemble W) (f : formula) :
  forall (M:Model A) (w:W), sat A M w f <-> sat A M w (NNF f).
Proof.
  functional induction (NNF f) using NNF_ind.
  - simpl; split; intro; assumption.
  - simpl; split; intro; assumption.
  - simpl; split; intro; assumption.
  - simpl; split; intro; assumption.
  - simpl; split; intro.
    + apply I. 
    + intro H0. assumption.
  - simpl; split; intro.
    + apply H. apply I.
    + intro H0. assumption.
  - simpl; split; intro.
    + apply NNPP in H. apply IHf0. apply H.
    + intro. apply H0. apply IHf0. assumption.
  - simpl; split; intro. apply not_and_or in H. destruct H as [H0|H1].
    + left. apply IHf0. assumption.
    + right. apply IHf1. assumption.
    + intro H1. destruct H1 as [H2 H3]. destruct H as [H4|H5].
      * apply IHf0 in H4. simpl in H4. apply H4. assumption.
      * apply IHf1 in H3. simpl in H3. apply H3. assumption. 
  - simpl; split; intro.
    + apply not_or_and in H. destruct H. split.
      * simpl in H. apply IHf0. assumption.
      * simpl in H. apply IHf1. assumption.
    + intro H0. destruct H. destruct H0.
      * apply IHf0 in H. simpl in H. apply H. assumption.
      * apply IHf1 in H1. simpl in H. apply H1. assumption.
  - simpl. split. intro. split. 
    + apply IHf0. apply not_or_and with (P:=(~sat A M w f9)) in H.
      destruct H. apply NNPP in H. assumption.
    + apply IHf1. apply not_or_and with (P:=(~sat A M w f9)) in H.
      destruct H. assumption.
    + intro. destruct H. apply IHf1 in H0.
      intro. simpl in H0. apply H0. destruct H1. apply IHf0 in H. contradiction. contradiction.
  - simpl; split; intro.
    + apply not_all_ex_not in H. destruct H as [x H0].
      exists x. apply imply_to_and in H0. unfold not in H0. destruct H0 as [H1 H2]. split.
    * assumption.
    * apply IHf0. assumption.
    + destruct H. intro. destruct H. apply H0 in H. apply IHf0 in H1.
      simpl in H1. contradiction.
  - simpl; split; intro. 
    + intros w' H0. apply not_ex_all_not with (n:=w') in H.
      apply IHf0. intro H1. unfold not in H. apply not_and_or in H.
      destruct H as [H2|H3]. apply H2. assumption. apply H3. assumption.
    + intro H0. destruct H0 as [x H1]. destruct H1. apply H in H0. apply IHf0 in H0.
      apply H0. assumption.
  - simpl; split; intro; destruct H as [H0 H1]; split; 
    try (apply IHf0); try (apply IHf1); assumption.
  - simpl; split; intro; destruct H as [H0|H1].
    + left. apply IHf0. assumption.
    + right. apply IHf1. assumption.
    + left. apply IHf0. assumption.
    + right. apply IHf1. assumption.
  - simpl; split; intro.
    + destruct H.
      * left. apply IHf0. simpl. assumption.
      * right. apply IHf1. assumption.
    + destruct H.
      * left. apply IHf0 in H. simpl in H. assumption.
      * right. apply IHf1 in H. assumption.
  - simpl; split; intro; intros w' H1; apply IHf0; apply H; assumption. 
  - simpl; split; intro; destruct H as [x H0]; exists x; destruct H0 as [H1 H2]; split;
    try assumption; try (apply IHf0; assumption).
Qed.

