Require Import Init.Nat.
Require Import Psatz.
Require Import Coq.Arith.PeanoNat.
Require Import Coq.Lists.List.

(* Authors: Cláudia Nalon & Maria Júlia Dias Lima *)

(** The set of well-formed formulae of the monomodal logic K is built from an enumerable set of propositional symbols: *)

Definition Props := nat.

(** and a set of constants where tt represent verum and ff represents falsum.: *)

Inductive Const: Set := tt | ff.

(** A variable symbolyzing the size of the agent set: **)

Variable n : nat.

(** The set of agents parametrized by the size, containg all natural numbers from 1 to n **)

Definition Agents :=
  {a:nat | (a <= n)}.

(** Function that turns an agent in the set above into a nat **)

Definition Agents_to_Nat (a:Agents) : nat :=
match a with
| exist _ m _ => m
end.

(** **)

Coercion Agents_to_Nat: Agents >-> nat.

(** The set of formulae is inductively defined as follows: **)

Inductive formula : Type :=
  | PropsF : Props -> formula
  | ConstF : Const -> formula
  | Not : formula -> formula
  | And : formula -> formula -> formula
  | Or : formula -> formula -> formula
  | Imp : formula -> formula -> formula
  | Box : Agents -> formula -> formula
  | Diamond : Agents -> formula -> formula.

(** The size of a formula is defined in the usual way: **)

Fixpoint size (f: formula) : nat :=
  match f with
  | PropsF p => 1
  | ConstF _ => 1
  | Not g => 1 + size g
  | And f g => 1 + (size f) + (size g)
  | Or f g => 1 + (size f) + (size g)
  | Imp f g => 1 + (size f) + (size g)
  | Box a f => 1 + (size f)
  | Diamond a f => 1 + (size f)
  end.

Inductive tree: Type :=
  | leaf (f:formula) (n:nat)
  | binary_op (f:formula) (n:nat) (t1 t2: tree)
  | unary_op (f:formula) (n:nat) (t1: tree).

Fixpoint insert (v:formula) (n:nat) : tree :=
  match v with
  | PropsF p => leaf (PropsF p) n
  | ConstF ff => leaf (ConstF ff) n
  | ConstF tt => leaf (ConstF tt) n
  | Not p => unary_op (Not p) n (insert p n)
  | And p q => binary_op (And p q) n (insert p n) (insert q n)
  | Or p q => binary_op (Or p q) n (insert p n)  (insert q n)
  | Imp p q => binary_op (Imp p q) n (insert p n) (insert q n)
  | Box a p => unary_op (Box a p) n (insert p (add 1 n))
  | Diamond a p => unary_op (Diamond a p) n (insert p (add 1 n))
  end.

Fixpoint return_ml (t:tree) : nat :=
  match t with
  | leaf _ n => n
  | unary_op _ n f' => (return_ml f')
  | binary_op _ n f' f''  => PeanoNat.Nat.max (return_ml f') (return_ml f'')
  end.

(** The maximum number of modal operators in a formula: **)

Fixpoint mdepth (f:formula) : nat :=
  match f with 
  | PropsF p => 0
  | ConstF _ => 0
  | Not g => (mdepth g)
  | And h g => PeanoNat.Nat.max(mdepth h) (mdepth g)
  | Or h g => PeanoNat.Nat.max(mdepth h) (mdepth g)
  | Imp h g => PeanoNat.Nat.max(mdepth h) (mdepth g)
  | Box a g => S (mdepth g)
  | Diamond a g => S (mdepth g)
  end.

(** List of propositional letters in a formula:  **)


Theorem mdepth_s (f:formula): 
  S (mdepth f + n) = mdepth f + (S n).
Proof.
  simpl. destruct n.
  - auto.
  - simpl. auto.
Qed.

Theorem n_0 (m:nat): 
 m = m+0.
Proof.
  auto.
Qed. 

Theorem S_return_ml (f:formula) (m:nat):
  (S (return_ml (insert f m)) = return_ml (insert f (S m))).
Proof.
  Admitted.


Theorem ml_is_mdepth (f:formula) :
  (mdepth f) + n = (return_ml (insert f n)).
Proof.
simpl. induction f.
- simpl. auto.
- simpl. destruct c; simpl; auto.
- simpl. auto.
- simpl. rewrite <- IHf1. rewrite  <- IHf2. lia.
- simpl. rewrite <- IHf1. rewrite  <- IHf2. lia.
- simpl. rewrite <- IHf1. rewrite  <- IHf2. lia.
- simpl. rewrite IHf. apply S_return_ml.
- simpl. rewrite IHf. apply S_return_ml.
Admitted.
